/*jslint browser: true, unparam: true, todo: true, evil: true*/
/*globals define: false, requestAnimationFrame: false, Event: false*/
define(['../util.js'], function (util) {
    'use strict';
    var components = {};
    // components that contain other components must be described here because
    // of the inherit circular reference to the components module
    function inputComponentFactory(inputOptions) {
        return function (header, index, pivotForm) {
            var component = util.createElement('span', null, {className: 'component input-type-' + inputOptions.type}),
                label = util.createElement('span', component, {className: 'label'}),
                input = util.createElement('input', component, inputOptions);
            label.innerHTML = header.title === undefined ? header.name : header.title;
            component.label = label;
            component.input = input;
            util.addEvents(input, header.events);
            util.addEvents(label, header.labelEvents);
            util.setProperties(input.style, header.style);
            util.setProperties(label.style, header.labelStyle);
            util.setProperties(component.containerStyle, header.containerStyle);
            component.addEventListener = input.addEventListener;
            component.removeEventListener = input.removeEventListener;
            return component;
        };
    }
    // html 4 input types
    ['checkbox', 'file', 'hidden', 'image', 'password', 'radio', 'reset', 'submit', 'text',
        // html 5 input types
        'color', 'date', 'datetime-local', 'email', 'month', 'number', 'range', 'search', 'tel', 'time',
        'url', 'week'].forEach(function (typeKey) {
        components[typeKey] = inputComponentFactory({className: 'input', type: typeKey});
    });
    components.select = function (header, index, pivotForm) {
        var componentContext = this,
            component = util.createElement('span', null, {className: 'component'}),
            label = util.createElement('span', component, {className: 'label'}),
            input = util.createElement('select', component, {className: 'select'}),
            fEnumVal,
            options = [];
        function fillOptions(arr) {
            arr.forEach(function (item) {
                var option;
                if (!Array.isArray(item)) { item = [item, item]; }
                option = util.createElement('option', input, {value: item[0], innerHTML: item[1]});
                util.setProperties(option, header.optionStyle);
                options.push(option);
            });
        }
        label.innerHTML = header.title === undefined ? header.name : header.title;
        component.label = label;
        component.input = input;
        component.containerStyle = {};
        util.addEvents(input, header.events);
        util.addEvents(label, header.labelEvents);
        util.setProperties(input.style, header.style);
        util.setProperties(label.style, header.labelStyle);
        util.setProperties(component.containerStyle, header.containerStyle);
        if (Array.isArray(header.enum)) {
            fillOptions(header.enum);
        } else if (typeof header.enum === 'function') {
            fEnumVal = header.enum.apply(componentContext, [function (value) {
                if (fEnumVal) { throw new Error('Async return detected from value already set via sync function.'); }
                fillOptions(value);
            }]);
            if (fEnumVal) {
                fillOptions(fEnumVal);
            }
        }
        component.addEventListener = input.addEventListener;
        component.removeEventListener = input.removeEventListener;
        return component;
    };
    components.HTMLElement = function (header, index, pivotForm) {
        var component = util.createElement(header.tagName, null, header);
        util.addEvents(component, header.componentEvents);
        util.setProperties(component.style, header.componentStyle);
        util.setProperties(component.containerStyle, header.containerStyle);
        return component;
    };
    ['h1', 'h2', 'h3', 'h4', 'button', 'hr', 'pre',
        'label', 'div', 'p', 'span', 'i', 'a'].forEach(function (h) {
        components[h] = function (header, index, pivotForm) {
            header.tagName = h;
            return components.HTMLElement(header, index, pivotForm);
        };
    });
    components.string = components.text;
    components['canvas-datagrid'] = function (header, index, pivotForm) {
        var pContext, component, grid, importScript;
        if (!window.canvasDatagrid) {
            importScript = util.createElement('script', null);
            importScript.src = header.src || 'https://unpkg.com/canvas-datagrid';
            importScript.type = 'text/javascript';
            importScript.async = false;
            document.getElementsByTagName('head')[0].appendChild(importScript);
        }
        pContext = this;
        component = util.createElement('div', null, {className: 'canvas-datagrid-child'});
        grid = window.canvasDatagrid();
        component.containerStyle = {
            width: '100%',
            height: '100%'
        };
        component.input = grid;
        header.static = header.static === undefined ? true : header.static;
        component.appendChild(grid);
        util.setProperties(grid, header);
        util.setProperties(component.containerStyle, header.containerStyle);
        grid.name = pContext.name ? (pContext.name + '_canvas-datagrid_' + index) : undefined;
        util.addEvents(grid, header.events);
        util.addEvents(component, header.componentEvents);
        pivotForm.addEventListener('resize', function () {
            component.resize();
        });
        component.addEventListener('resize', function () {
            component.resize();
        });
        component.resize = function () {
            setTimeout(function () { grid.resize(true); }, 10);
        };
        function changeEvent() {
            if (!pivotForm.data) { return; }
            pivotForm.dispatchEvent(new Event('change'));
        }
        grid.addEventListener('endedit', changeEvent);
        Object.defineProperty(grid, 'value', {
            set: function (value) {
                grid.data = value;
            },
            get: function () {
                return grid.data;
            }
        });
        return component;
    };
    // hierarchical components
    components['pivot-form'] = function (header, index, pivotForm) {
        var component;
        component = window.pivotForm();
        component.containerStyle = {};
        component.className = 'pivot-form-child';
        pivotForm.addChildForm(component);
        component.index = index;
        header.static = header.static === undefined ? true : header.static;
        util.setProperties(component.style, header.style);
        util.addEvents(component, header.events);
        Object.defineProperty(component, 'value', {
            get: function () {
                return component.data;
            },
            set: function (value) {
                component.data = value;
            }
        });
        return component;
    };
    components.cutter = function (header, index, pivotForm) {
        var pContext = this,
            component = util.createElement('div', null, {className: 'cutter-component'}),
            cutter = util.createElement('div', null, {className: 'cutter-cutter'}),
            cutterWidth = 3,
            margin = 10,
            position,
            moveStart,
            orientation = 'vertical',
            cursor = {
                vertical: 'ew-resize',
                horizontal: 'ns-resize'
            },
            cutterSize = {
                vertical: 'minHeight',
                horizontal: 'minWidth'
            },
            size = {
                vertical: 'minWidth',
                horizontal: 'minHeight'
            },
            offset = {
                vertical: 'offsetWidth',
                horizontal: 'offsetHeight',
            };
        component.cutter = cutter;
        cutter.style.cursor = cursor[orientation];
        Object.defineProperty(component, 'position', {
            get: function () {
                return position;
            },
            set: function (value) {
                if (isNaN(value)) { throw new TypeError('position must be a number.'); }
                if (value < 0) { value = 0; }
                if (value > component[offset[orientation]]) { value = 0; }
                position = value;
                component.resize();
            }
        });
        Object.defineProperty(component, 'orientation', {
            get: function () {
                return orientation;
            },
            set: function (value) {
                if (!(value === 'vertical' || value === 'horizontal')) {
                    throw new TypeError('orientation must be vertical or horizontal.');
                }
                orientation = value;
                cutter.style.cursor = cursor[value];
                component.resize();
            }
        });
        function moveCutter(e) {
            var offsetDelta = {
                minWidth: moveStart.x - e.clientX,
                minHeight: moveStart.y - e.clientY,
            };
            component.position = moveStart.p - offsetDelta[size[orientation]];
        }
        function endMoveCutter(e) {
            document.body.removeEventListener('mousemove', moveCutter);
            document.body.removeEventListener('mouseup', endMoveCutter);
        }
        cutter.onmousedown = function (e) {
            moveStart = {
                x: e.clientX,
                y: e.clientY,
                p: position
            };
            document.body.addEventListener('mouseup', endMoveCutter);
            document.body.addEventListener('mousemove', moveCutter);
        };
        component.resize = function () {
            var s = component[offset[orientation]];
            if (position === undefined) {
                position = s / 2;
            }
            cutter.style[cutterSize[orientation]] = '100%';
            cutter.style[size[orientation]] = cutterWidth + 'px';
            component.panels[0].style[size[orientation]] = position - margin + 'px';
            component.panels[1].style[size[orientation]] = margin + s - position - (margin * 0.5) + 'px';
            [0, 1].forEach(function (n) {
                var panel = component.panels[n];
                if (panel.dispatchEvent) {
                    panel.dispatchEvent(new Event('resize'));
                }
            });
        };
        pivotForm.addEventListener('resize', component.resize);
        component.isContainer = true;
        component.panels = [];
        component.containerStyle = {
            width: '100%'
        };
        [0, 1].forEach(function (n) {
            var panel;
            if (header.panels[n]) {
                panel = components['pivot-form'].call(pContext, header.panels[n], index, pivotForm);
                panel.classList.add('cutter-panel');
                panel.classList.add('cutter-panel-' + n);
                util.setProperties(panel, header.panels[n]);
                util.addEvents(panel, header.events);
                component.appendChild(panel);
                if (n === 0) {
                    component.appendChild(cutter);
                }
                component.panels.push(panel);
            }
        });
        setTimeout(function () {
            position = component[offset[orientation]] / 2;
            component.resize();
        }, 0);
        return component;
    };
    components.tabs = function (header, index, pivotForm) {
        var pContext = this,
            component = util.createElement('div', null, {className: 'tabbed-component'}),
            tabsContainer = util.createElement('div', component, {className: 'tabs'}),
            activeTabIndex,
            tabs = [],
            defaultTab = header.defaultTabIndex || 0;
        component.containerStyle = {
            width: '100%'
        };
        component.isContainer = true;
        component.tabsContainer = tabsContainer;
        component.tabs = tabs;
        header.static = header.static === undefined ? true : header.static;
        function activateTab(tabIndex) {
            var activeTabItem = tabs[tabIndex],
                e = new Event('activatetab');
            e.tab = activeTabItem;
            if (!activeTabItem) { return; }
            activeTabIndex = tabIndex;
            tabs.forEach(function (tab) {
                var ev = new Event('deactivatetab');
                ev.tab = tab;
                tab.tabElement.classList.add('tab-inactive');
                tab.tabElement.classList.remove('tab-active');
                tab.content.classList.remove('tab-content-active');
                if (tab.content.parentNode) {
                    tab.content.parentNode.removeChild(tab.content);
                    component.dispatchEvent(ev);
                }
            });
            activeTabItem.tabElement.classList.add('tab-active');
            activeTabItem.tabElement.classList.remove('tab-inactive');
            activeTabItem.content.classList.add('tab-content-active');
            component.appendChild(activeTabItem.content);
            component.dispatchEvent(e);
        }
        header.tabs.forEach(function (tabHeader, index) {
            var tab = {};
            tab.tabElement = util.createElement('div', tabsContainer, {
                className: 'tab ' + 'tab-' + util.varSafe(tabHeader.name),
                events: {
                    click: function (e) {
                        if (activeTabIndex === index) { return; }
                        activateTab(index);
                    }
                },
                innerHTML: tabHeader.title || tabHeader.name
            });
            tab.content = components['pivot-form'].call(pContext, tabHeader, index, pivotForm);
            pivotForm.addEventListener('resize', function () {
                tab.content.dispatchEvent(new Event('resize'));
            });
            tab.content.classList.add('tab-content');
            util.setProperties(tab.content, tabHeader);
            util.setProperties(tab.tabElement, tabHeader.tabStyle);
            util.addEvents(tab.tabElement, tabHeader.tabEvents);
            util.addEvents(tab.content, tabHeader.contentEvents);
            util.addEvents(tab.content, tabHeader.events);
            tabs.push(tab);
        });
        util.addEvents(component, header.events);
        util.setProperties(component.containerStyle, header.containerStyle);
        // instantiate each tab
        header.tabs.forEach(function (tabName, index) {
            if (defaultTab === index) { return; }
            requestAnimationFrame(function () { activateTab(index); });
        });
        requestAnimationFrame(function () { activateTab(defaultTab || 0); });
        Object.defineProperty(component, 'tabIndex', {
            get: function () {
                return activeTabIndex;
            },
            set: function (value) {
                return activateTab(Math.max(value, tabs.length - 1));
            }
        });
        Object.defineProperty(component, 'value', {
            get: function () {
                var d = {};
                tabs.forEach(function (tab) {
                    Object.assign(d, tab.data);
                });
                return d;
            },
            set: function (value) {
                tabs.forEach(function (tab) {
                    tab.value = value;
                });
                return;
            }
        });
        return component;
    };
    return components;
});
